/** 
 * 
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/
package org.codehaus.activecluster.impl;

import EDU.oswego.cs.dl.util.concurrent.SynchronizedBoolean;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.codehaus.activecluster.Cluster;
import org.codehaus.activecluster.ClusterListener;
import org.codehaus.activecluster.LocalNode;
import org.codehaus.activecluster.Service;

import javax.jms.*;
import java.io.Serializable;
import java.util.Map;
import java.util.Timer;

/**
 * A default implementation of ActiveCluster which uses standard JMS operations
 *
 * @version $Revision: 1.12 $
 */
public class DefaultCluster implements Cluster {

    private final static Log log = LogFactory.getLog(DefaultCluster.class);

    private StateServiceImpl stateService;
    private LocalNode localNode;
    private Topic destination;
    private Connection connection;
    private Session session;
    private MessageProducer producer;
    private MessageConsumer consumer;
    private Timer timer;
    private SynchronizedBoolean started = new SynchronizedBoolean(false);
    private Object clusterLock = new Object();

    public DefaultCluster(final LocalNode localNode, Topic dataTopic, Topic destination, Connection connection, Session session,
                          MessageProducer producer, Timer timer, long inactiveTime) throws JMSException {
        this.localNode = localNode;
        this.destination = destination;
        this.connection = connection;
        this.session = session;
        this.producer = producer;
        this.timer = timer;

        if (producer == null) {
            throw new IllegalArgumentException("No producer specified!");
        }

        // now lets subscribe the service to the updates from the data topic
        consumer = session.createConsumer(dataTopic, null, true);

        log.info("Creating data consumer on topic: " + dataTopic);

        this.stateService = new StateServiceImpl(this, clusterLock, new Runnable() {
            public void run() {
                if (localNode instanceof ReplicatedLocalNode) {
                    ((ReplicatedLocalNode) localNode).pingRemoteNodes();
                }
            }
        }, timer, inactiveTime);
        consumer.setMessageListener(new StateConsumer(stateService));
    }

    public synchronized void addClusterListener(ClusterListener listener) {
        stateService.addClusterListener(listener);
    }

    public synchronized void removeClusterListener(ClusterListener listener) {
        stateService.removeClusterListener(listener);
    }

    public Topic getDestination() {
        return destination;
    }

    public LocalNode getLocalNode() {
        return localNode;
    }

    public Map getNodes() {
        return stateService.getNodes();
    }

    public synchronized void send(Destination destination, Message message) throws JMSException {
        producer.send(destination, message);
    }

    public synchronized MessageConsumer createConsumer(Destination destination) throws JMSException {
        return getSession().createConsumer(destination);
    }

    public MessageConsumer createConsumer(Destination destination, String selector) throws JMSException {
        return getSession().createConsumer(destination, selector);
    }

    public MessageConsumer createConsumer(Destination destination, String selector, boolean noLocal) throws JMSException {
        return getSession().createConsumer(destination, selector, noLocal);
    }

    public synchronized Message createMessage() throws JMSException {
        return getSession().createMessage();
    }

    public synchronized BytesMessage createBytesMessage() throws JMSException {
        return getSession().createBytesMessage();
    }

    public synchronized MapMessage createMapMessage() throws JMSException {
        return getSession().createMapMessage();
    }

    public synchronized ObjectMessage createObjectMessage() throws JMSException {
        return getSession().createObjectMessage();
    }

    public synchronized ObjectMessage createObjectMessage(Serializable object) throws JMSException {
        return getSession().createObjectMessage(object);
    }

    public synchronized StreamMessage createStreamMessage() throws JMSException {
        return getSession().createStreamMessage();
    }

    public synchronized TextMessage createTextMessage() throws JMSException {
        return getSession().createTextMessage();
    }

    public synchronized TextMessage createTextMessage(String text) throws JMSException {
        return getSession().createTextMessage(text);
    }

    public synchronized void start() throws JMSException {
        if (started.commit(false, true)) {
            connection.start();
        }
    }

    public void stop() throws JMSException {
        try {
            if (localNode instanceof Service) {
                ((Service) localNode).stop();
            }
            timer.cancel();
            session.close();
            connection.stop();
            connection.close();
        }
        finally {
            connection = null;
            session = null;
        }
    }

    public boolean waitForClusterToComplete(int expectedCount, long timeout) throws InterruptedException {
        timeout = timeout > 0 ? timeout : Long.MAX_VALUE;
        long waitTime = timeout;
        long start = System.currentTimeMillis();
        synchronized (clusterLock) {
            while (stateService.getNodes().size() < expectedCount && started.get() && waitTime > 0) {
                clusterLock.wait(waitTime);
                waitTime = timeout - (System.currentTimeMillis() - start);
            }
        }
        return stateService.getNodes().size() >= expectedCount;
    }

    protected Session getSession() throws JMSException {
        if (session == null) {
            throw new JMSException("Cannot perform operation, this cluster connection is now closed");
        }
        return session;
    }
}
