/**
 * 
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/
package org.codehaus.activecluster.impl;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.codehaus.activecluster.Cluster;
import org.codehaus.activecluster.ClusterException;
import org.codehaus.activecluster.ClusterFactory;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.jms.Topic;
import javax.jms.DeliveryMode;
import java.util.Timer;

/**
 * A Factory of DefaultCluster instances
 *
 * @version $Revision: 1.12 $
 */
public class DefaultClusterFactory implements ClusterFactory {

    private final static Log log = LogFactory.getLog(DefaultClusterFactory.class);

    private ConnectionFactory connectionFactory;
    private boolean transacted;
    private int acknowledgeMode;
    private String dataTopicPrefix;
    private long inactiveTime;
    private boolean useQueueForInbox = false;
    private int deliveryMode = DeliveryMode.NON_PERSISTENT;

    public DefaultClusterFactory(ConnectionFactory connectionFactory, boolean transacted, int acknowledgeMode, String dataTopicPrefix, long inactiveTime) {
        this.connectionFactory = connectionFactory;
        this.transacted = transacted;
        this.acknowledgeMode = acknowledgeMode;
        this.dataTopicPrefix = dataTopicPrefix;
        this.inactiveTime = inactiveTime;
    }

    public DefaultClusterFactory(ConnectionFactory connectionFactory) {
        this(connectionFactory, false, Session.AUTO_ACKNOWLEDGE, "ACTIVECLUSTER.DATA.", 5000L);
    }

    public Cluster createCluster(Topic groupDestination) throws ClusterException, JMSException {
        Connection connection = getConnectionFactory().createConnection();
        Session session = createSession(connection);
        return createCluster(connection, session, groupDestination);
    }

    public Cluster createCluster(String topic) throws ClusterException, JMSException {
        Connection connection = getConnectionFactory().createConnection();
        Session session = createSession(connection);
        Topic groupDestination = session.createTopic(topic);
        return createCluster(connection, session, groupDestination);
    }

    // Properties
    //-------------------------------------------------------------------------
    public String getDataTopicPrefix() {
        return dataTopicPrefix;
    }

    public void setDataTopicPrefix(String dataTopicPrefix) {
        this.dataTopicPrefix = dataTopicPrefix;
    }

    public int getAcknowledgeMode() {
        return acknowledgeMode;
    }

    public void setAcknowledgeMode(int acknowledgeMode) {
        this.acknowledgeMode = acknowledgeMode;
    }

    public long getInactiveTime() {
        return inactiveTime;
    }

    public void setInactiveTime(long inactiveTime) {
        this.inactiveTime = inactiveTime;
    }

    public boolean isTransacted() {
        return transacted;
    }

    public void setTransacted(boolean transacted) {
        this.transacted = transacted;
    }

    public boolean isUseQueueForInbox() {
        return useQueueForInbox;
    }

    public void setUseQueueForInbox(boolean useQueueForInbox) {
        this.useQueueForInbox = useQueueForInbox;
    }

    public ConnectionFactory getConnectionFactory() {
        return connectionFactory;
    }

    public void setConnectionFactory(ConnectionFactory connectionFactory) {
        this.connectionFactory = connectionFactory;
    }

    public int getDeliveryMode() {
        return deliveryMode;
    }

    /**
     * Sets the delivery mode of the group based producer
     */
    public void setDeliveryMode(int deliveryMode) {
        this.deliveryMode = deliveryMode;
    }

    // Implementation methods
    //-------------------------------------------------------------------------
    protected Cluster createCluster(Connection connection, Session session, Topic groupDestination) throws JMSException {
        Topic dataTopic = session.createTopic(dataTopicPrefix + groupDestination.getTopicName());

        log.info("Creating cluster group producer on topic: " + groupDestination);

        MessageProducer producer = createProducer(session, null);
        producer.setDeliveryMode(deliveryMode);

        log.info("Creating cluster data producer on topic: " + dataTopic);

        MessageProducer keepAliveProducer = session.createProducer(dataTopic);
        keepAliveProducer.setDeliveryMode(DeliveryMode.NON_PERSISTENT);
        StateService serviceStub = new StateServiceStub(session, keepAliveProducer);

        Destination localInbox = null;
        if (useQueueForInbox) {
            localInbox = session.createTemporaryQueue();
        }
        else {
            localInbox = session.createTemporaryTopic();
        }
        ReplicatedLocalNode localNode = new ReplicatedLocalNode(localInbox, serviceStub);
        Timer timer = new Timer();
        DefaultCluster answer = new DefaultCluster(localNode, dataTopic, groupDestination, connection, session, producer, timer, inactiveTime);
        return answer;
    }

    /*
     protected Cluster createInternalCluster(Session session, Topic dataDestination) {
         MessageProducer producer = createProducer(session);
         return new DefaultCluster(new NonReplicatedLocalNode(), dataDestination, connection, session, producer);
     }
     */
    
    protected MessageProducer createProducer(Session session, Topic groupDestination) throws JMSException {
        return session.createProducer(groupDestination);
    }

    protected Session createSession(Connection connection) throws JMSException {
        return connection.createSession(transacted, acknowledgeMode);
    }
}